import logging
import time

import jax
import jax.numpy as jnp
import numpy as np
import matplotlib.pyplot as plt

import pyfstat
from fasttracks.dataio import SFTDataIO
from fasttracks import likelihood, tracks

logging.basicConfig(
    level=logging.INFO,
    format=f"%(asctime)s.%(msecs)03d %(name)s %(levelname)-8s: %(message)s",
    datefmt="%y-%m-%d %H:%M:%S",
)

# Create injection
writer_kwargs = {
    "label": "binaryAllSkyInjection",
    "outdir": "binaryAllSkyInjection",
    "tstart": 1368931380,
    "duration": 86400 * 30 * 1,
    "detectors": "H1,L1",
    "sqrtSX": 1e-23,
    "Tsft": 1800,
    "SFTWindowType": "tukey",
    "SFTWindowParam": 0.01,
    "Band": 2.0,
}

signal_kwargs = {
    "F0": 100.0,
    "F1": 0,
    "asini": 10.0,
    "period": 864000.0,
    "tp": writer_kwargs["tstart"] + 0.5 * writer_kwargs["duration"],
    "h0": writer_kwargs["sqrtSX"] / 5.0,
    "psi": 0,
    "cosi": 1,
    "phi": 0,
    "Alpha": 0.0,
    "Delta": 0.0,
}
writer = pyfstat.BinaryModulatedWriter(**writer_kwargs, **signal_kwargs)
writer.make_data()

targeted_theta = jnp.array(
    [signal_kwargs[key] for key in ["F0", "Alpha", "Delta", "asini", "period", "tp"]]
)
targeted_theta = targeted_theta.at[2].set(jnp.sin(targeted_theta[2]))
targeted_theta = targeted_theta.at[4].set(2 * jnp.pi / targeted_theta[4])
targeted_theta = targeted_theta.at[5].set(
    targeted_theta[4] * targeted_theta[5]
)  # phib = Omega * tasc

dvalues = jnp.zeros(len(targeted_theta))
dvalues = dvalues.at[3].set(2)
dvalues = dvalues.at[4].set(1e-6)
dvalues = dvalues.at[5].set(jnp.pi)

upper_vals = targeted_theta + dvalues
lower_vals = targeted_theta - dvalues


def get_batch(key):
    key, subkey = jax.random.split(key)
    return key, jax.random.uniform(
        subkey,
        shape=(batch_size, 6),
        minval=lower_vals,
        maxval=upper_vals,
    )


# Compute detection statistics
power, timestamps, velocities, weights, t_sft, bin_0 = SFTDataIO(
    sftfilepath=writer.sftfilepath,
    freq_min=99.95,
    freq_max=100.05,
)(sky_position=(signal_kwargs["Alpha"], signal_kwargs["Delta"]))


def statistic(template):
    track_hertz = tracks.f_of_t_circular_binary(template, timestamps, velocities)
    return likelihood.normalized_power(track_hertz, weights, power, t_sft, bin_0)


fast_statistic = jax.jit(jax.vmap(statistic, in_axes=0, out_axes=0))

key = jax.random.PRNGKey(0)
batch_size = 50000
num_batches = 10
num_templates = batch_size * num_batches

templates = jnp.zeros((num_templates, 6))
powers = jnp.zeros(num_templates)

logging.info(f"Num. templates: {num_templates}")
logging.info("Starting compute statistic for loop")

for ind in range(num_batches):
    key, batch_thetas = get_batch(key)

    t0 = time.time()
    batch_power = fast_statistic(batch_thetas).block_until_ready()
    logging.info(f"Time to run a batch: {time.time() - t0:.2g} s")

    left = ind * batch_size
    right = left + batch_size

    templates = templates.at[left:right].set(batch_thetas)
    powers = powers.at[left:right].set(batch_power)

logging.info("Done!")

logging.info(f"Computing targeted track power")
targeted_track = tracks.f_of_t_circular_binary(targeted_theta, timestamps, velocities)
targeted_power = likelihood.normalized_power(
    targeted_track, weights, power, t_sft, bin_0
)

# Make a plot of the targeted_power results
logging.info(f"Generating templates vs power plot")
sort_ind = jnp.argsort(powers)
stat = powers[sort_ind]
aps = templates[:, 3][sort_ind]
pds = 2 * jnp.pi / templates[:, 4][sort_ind]

fig, axs = plt.subplots(2, 2, figsize=(10, 8))
scattering = axs[1, 0].scatter(pds, aps, c=stat, cmap="coolwarm")
axs[0, 0].scatter(pds, stat, c=stat, cmap="coolwarm")
axs[0, 0].scatter(signal_kwargs["period"], targeted_power, c="r", marker="*")
axs[1, 1].scatter(stat, aps, c=stat, cmap="coolwarm")
axs[1, 1].scatter(targeted_power, signal_kwargs["asini"], c="r", marker="*")

axs[1, 0].set(
    xlabel="Orbital Period (s)",
    ylabel="Projected Semimajor Axis (s)",
)
axs[1, 1].set(
    ylabel="Projected Semimajor Axis (s)",
    xlabel="Weighed Normalized Power",
)
axs[0, 0].set(
    xlabel="Orbital Period (s)",
    ylabel="Weighed Normalized Power",
)
axs[0, 0].axvline(signal_kwargs["period"], ls="--", c="k", lw=1, alpha=0.9)
axs[1, 0].axhline(signal_kwargs["asini"], ls="--", c="k", lw=1, alpha=0.9)
axs[1, 0].axvline(signal_kwargs["period"], ls="--", c="k", lw=1, alpha=0.9)
axs[1, 1].axhline(signal_kwargs["asini"], ls="--", c="k", lw=1, alpha=0.9)
axs[0, 1].axis("off")
for ax in axs.flat:
    ax.grid(which="major", ls="-", alpha=0.2)
    ax.grid(which="minor", ls="-", alpha=0.1)
    ax.minorticks_on()
fig.colorbar(scattering, label="power", ax=axs[0, 1], location="left")
fig.savefig(f"{writer_kwargs['outdir']}/batch_power.png")

# Make a plot of the generated data
# This concatenates all detectors, it is what it is.
logging.info(f"Generating spectrogram plot")
fig, ax = plt.subplots(figsize=(16, 10))
c = ax.pcolormesh(power)
targeted_track_bins = (targeted_track * t_sft + 0.5).astype("uint32") - bin_0
ax.plot(
    targeted_track_bins + 1,
    color="white",
    label="Injected Signal",
    lw=0.5,
)
ax.plot(
    targeted_track_bins - 1,
    color="white",
    lw=0.5,
)
fig.colorbar(c)
fig.tight_layout()
fig.savefig(f"{writer_kwargs['outdir']}/spectrogram.png")

logging.info(f"All done")



import jax.numpy as jnp

def normalized_power(track_hertz, weights, power, t_sft, bin_0):
    track_bins = (track_hertz * t_sft + 0.5).astype("uint32") - bin_0
    return jnp.sum(weights * power[track_bins,  jnp.arange(track_hertz.shape[0])])



import jax
import jax.numpy as jnp
import pytest

from fasttracks import tracks


@pytest.fixture
def timestamps():
    return jnp.arange(1000)


@pytest.fixture
def velocities(timestamps):
    return jnp.ones((3, timestamps.shape[0]))


@pytest.fixture
def theta_isolated():
    return jnp.array([100.0, 0.0, 0.0, 0.0, 0.0])


@pytest.fixture
def batch_isolated():
    return jnp.vstack([jnp.array([100.0, 0.0, 0.0, 0.0, 0.0]) for i in range(10)])


def test_isolated_track(theta_isolated, timestamps, velocities):
    tracks.f_of_t_isolated(theta_isolated, timestamps, velocities)

def test_vmap_track(batch_isolated, timestamps, velocities):
    batch_out = jax.vmap(tracks.f_of_t_isolated, in_axes=(0, None, None), out_axes=0)(
        batch_isolated, timestamps, velocities
    )
    print(batch_out)
    print(batch_out.shape)
    assert batch_out.shape[0] == batch_isolated.shape[0]



import os
from shutil import rmtree
import sys

import pytest

import pyfstat
from fasttracks.dataio import SFTDataIO

@pytest.fixture
def writer_kwargs(tmp_path):
    return {
        "label": "anInjection",
        "outdir": tmp_path / "SampleData",
        "tstart": 1238166018,
        "duration": 10 * 86400,
        "detectors": "H1,L1",
        "sqrtSX": 1e-23,
        "Tsft": 1800,
        "SFTWindowType": "tukey",
        "SFTWindowParam": 0.01,
        "Band": 0.2,
    }

@pytest.fixture
def signal_parameters(writer_kwargs):
    return {
        "F0": 100.0,
        "F1": -1e-9,
        "Alpha": 0.0,
        "Delta": 0.0,
        "asini": 10.0,
        "period": 864000.0,
        "tp": writer_kwargs["tstart"] + 0.5 * writer_kwargs["duration"],
        "h0": 1e-22,
        "cosi": 1,
        "psi": 0.0,
        "phi": 0.0,
        "tref": writer_kwargs["tstart"],
    }

@pytest.fixture
def writer(writer_kwargs, signal_parameters):
    writer = pyfstat.BinaryModulatedWriter(**writer_kwargs, **signal_parameters)
    writer.make_data()
    yield writer
    rmtree(writer.outdir)

def test_read_sfts(writer):
    power, timestamps, velocities, weights, t_sft, bin_0 = SFTDataIO(
        sftfilepath=writer.sftfilepath,
        freq_min=99.95,
        freq_max=100.05,
    )(sky_position=(0, 0))

    assert velocities.shape[0] == 3
    assert power.shape[1] == timestamps.shape[0]
    assert timestamps.shape[0] == weights.shape[0]


import logging

import jax.numpy as jnp
import lal
import lalpulsar
import numpy as np
from pyfstat import DetectorStates
from pyfstat.utils import get_sft_as_arrays

logger = logging.getLogger(__name__)


class SFTDataIO:
    def __init__(
        self,
        sftfilepath: str,
        freq_min: float = -1,
        freq_max: float = -1,
        running_median_window: int = lalpulsar.FstatOptionalArgsDefaults.runningMedianWindow,
    ):
        """
        sftfilepath:
            Path to SFT data file(s).
        freq_min:
            Minimum SFT frequency to load. `-1` means the lowest frequency
            present on the file.
        freq_max:
            Minimum SFT frequency to load. `-1` means the highest frequency
            present on the file.
        running_median_window:
            Number of bins to compute a running median. Defaults to lalpulsar's default.
            Basically don't touch it unless you know what you are doing.
        """
        self.sftfilepath = sftfilepath
        self.freq_min = freq_min
        self.freq_max = freq_max
        self.running_median_window = running_median_window

        self._get_data_from_sfts()

    def __call__(self, sky_position: tuple[float, float] | None = None):
        """
        Returns SFT power, timestamps, velocities, and weights
        for a given sky position so that JAX can operate on them.

        Weights are normalized so that `weights.sum() == 1`.

        Parameters
        -------------
        sky_position:
            Tuple containing (right_ascension, declination) in radians.

        Returns
        ---------
        tuple[power, timestamps, velocities, weights, t_sft, bin_0]
        """
        power = jnp.hstack([val["power"] for val in self.multisft_data.values()])
        timestamps = jnp.hstack(
            [val["timestamps"] for val in self.multisft_data.values()]
        )
        velocities = jnp.hstack(
            [val["velocities"] for val in self.multisft_data.values()]
        )

        weights = self.get_weights(sky_position)
        weights = weights / weights.sum()

        return (
            power,
            timestamps,
            velocities,
            weights,
            jnp.array(
                [
                    self.common_data["t_sft"],
                ]
            ),
            jnp.array([self.common_data["bin_0"]]),
        )

    def get_weights(self, sky_position=None):
        weights = self.get_noise_weights()
        if sky_position is not None:
            a, b = self.get_a_b(sky_position)
            weights = (a**2 + b**2) * weights
        return weights

    def get_noise_weights(self):
        return jnp.hstack([val["noise_weights"] for val in self.multisft_data.values()])

    def get_a_b(self, sky_position):
        AMcoeffs = self._get_AMcoeffs(*sky_position)
        a = jnp.hstack([val.a.data for val in AMcoeffs.values()])
        b = jnp.hstack([val.b.data for val in AMcoeffs.values()])
        return a, b

    def _get_data_from_sfts(self):
        """
        Uses PyFstat's get_sfts_as_arrays function to get the SFT's
        frequencies, timestamps, power, and noise weights.
        """
        logger.info(f"Parsing SFT catalog from {self.sftfilepath}...")
        sft_catalog = lalpulsar.SFTdataFind(self.sftfilepath, None)

        ifo_labels = lalpulsar.ListIFOsInCatalog(sft_catalog)

        logger.info(f"Reading SFTs from {self.freq_min} Hz to {self.freq_max} Hz...")
        multi_sfts = lalpulsar.LoadMultiSFTs(
            sft_catalog,
            fMin=self.freq_min,
            fMax=self.freq_max,
        )
        assert multi_sfts.length == ifo_labels.length

        logger.info(f"Normalizing SFTs and computing noise weights...")
        rngmed = lalpulsar.NormalizeMultiSFTVect(
            multsft=multi_sfts,
            blockSize=self.running_median_window,
            assumeSqrtSX=None,
        )
        multi_noise_weights = lalpulsar.ComputeMultiNoiseWeights(
            rngmed, self.running_median_window, 0
        )
        assert multi_noise_weights.length == multi_sfts.length
        assert not multi_noise_weights.isNotNormalized

        multidetector_states = DetectorStates().get_multi_detector_states_from_sfts(
            self.sftfilepath,
            central_frequency=0.5 * (self.freq_min + self.freq_max),
            time_offset=0.5 / multi_sfts.data[0].data[0].deltaF,
        )
        assert multidetector_states.length == ifo_labels.length

        for ifo_ind in range(ifo_labels.length):
            assert (
                multi_sfts.data[ifo_ind].length
                == multi_noise_weights.data[ifo_ind].length
            )
            assert (
                multi_sfts.data[ifo_ind].length
                == multidetector_states.data[ifo_ind].length
            )

        self.multisft_data = {}
        for ifo_ind, ifo_name in enumerate(ifo_labels.data):
            logging.debug(f"Reading data from IFO {ifo_name}")

            sfts = multi_sfts.data[ifo_ind]
            sft_amplitude = np.array([sft.data.data for sft in sfts.data]).T
            f0 = sfts.data[0].f0
            df = sfts.data[0].deltaF

            sft_data = {}
            sft_data["timestamps"] = np.array(
                [sft.epoch.gpsSeconds for sft in sfts.data]
            )
            sft_data["power"] = 2.0 * (
                sft_amplitude.real**2 + sft_amplitude.imag**2
            )
            sft_data["noise_weights"] = multi_noise_weights.data[ifo_ind].data

            nbins, nsfts = sft_data["power"].shape

            sft_data["frequency_Hz"] = np.linspace(f0, f0 + (nbins - 1) * df, nbins)
            sft_data["t_sft"] = int(
                1 / (sft_data["frequency_Hz"][1] - sft_data["frequency_Hz"][0]) + 0.5
            )
            sft_data["bin_0"] = int(
                sft_data["t_sft"] * sft_data["frequency_Hz"][0] + 0.5
            )

            sft_data["states"] = multidetector_states.data[ifo_ind]
            sft_data["velocities"] = np.vstack(
                [data.vDetector for data in multidetector_states.data[ifo_ind].data]
            ).T

            self.multisft_data[ifo_name] = sft_data

        self.common_data = {}
        for common_key in ["t_sft", "bin_0"]:
            common_value = list(
                set([ifo_data[common_key] for ifo_data in self.multisft_data.values()])
            )
            assert len(common_value) == 1
            self.common_data[common_key] = common_value[0]

    def _get_AMcoeffs(self, right_ascension: float, declination: float):
        logging.debug(
            f"Computing antenna pattern functions at ({right_ascension}, {declination})"
        )
        skypos = lal.SkyPosition()
        skypos.longitude = right_ascension
        skypos.latitude = declination
        skypos.system = lal.COORDINATESYSTEM_EQUATORIAL
        lal.NormalizeSkyPosition(skypos.longitude, skypos.latitude)

        AMcoeffs = {}
        for ifo_ind, ifo_label in enumerate(self.multisft_data):
            AMcoeffs[ifo_label] = lalpulsar.ComputeAMCoeffs(
                DetectorStates=self.multisft_data[ifo_label]["states"],
                skypos=skypos,
            )
        return AMcoeffs



import logging

import jax.numpy as jnp

logger = logging.getLogger(__name__)


def equatorial_to_cartesian(right_ascension, sin_declination):
    # -pi/2 < delta < pi/2 --> cos(delta) > 0
    cos_declination = jnp.sqrt(1 - sin_declination**2)
    return jnp.array(
        [
            [
                cos_declination * jnp.cos(right_ascension),
                cos_declination * jnp.sin(right_ascension),
                sin_declination,
            ]
        ]
    )


def f_of_t_isolated(theta, timestamps, velocities):
    """
    Parameters
    -------------
    theta:
        (f0, f1, ra, sin_dec, tref)
    timestamps:
        (num_timestamps,)
    velocities:
        (3, num_timestamps)
    """
    cartesian = equatorial_to_cartesian(theta[2], theta[3])
    return (theta[0] + (timestamps - theta[-1]) * theta[1]) * (
        1 + jnp.dot(cartesian, velocities).squeeze()
    )


def f_of_t_circular_binary(theta, timestamps, velocities):
    """
    Parameters
    -------------
    theta:
        (f0, ra, sin_dec, ap, omega, phib)

    phib is defined as `phib = omega * tasc`
    """
    cartesian = equatorial_to_cartesian(theta[1], theta[2])
    return theta[0] * (
        1
        + jnp.dot(cartesian, velocities).squeeze()
        - theta[3] * theta[4] * jnp.cos(theta[4] * timestamps - theta[5])
    )
